'''
This is the "Shaft Profile" class.
This class is responsible for defining the properties of a Shaft Profile object, such as:

I) Given properties:
--> 1) "name": a string of characters acting as a label
--> 2) "radii": a numpy array representing the external radius of the shaft at each step along the shaft's axis
--> 3) "locs": a numpy array representing the locations of the specified raddi along the shaft's axis

II) Calculated properties:
--> 1) "shaft": a shaft object to which the shaft's profile belongs
--> 2) "Area:: a numpy array representing the area of the cross-section of the shaft at each step along the shaft's profile expressed in [mm^2]
--> 3) "Wb": a numpy array representing the bending section modulus of the cross-section of the shaft at each step along the shaft's profile expressed in [mm^3]
--> 4) "Wt": a numpy array representing the torsion section modulus of the cross-section of the shaft at each step along the shaft's profile expressed in [mm^3]
--> 5) "sigma_N": a numpy array representing the normal stress acting on the shaft at each step along the shaft's profile expressed in [MPa]
--> 6) "sigma_Mb": a numpy array representing the resultant bending stress acting on the shaft at each step along the shaft's profile expressed in [MPa]
--> 7) "tau_Mt": a numpy array representing the torsional stress acting on the shaft at each step along the shaft's profile expressed in [MPa]
--> 8) "sigma_tot": a numpy array representing the equivalent normal stress acting on the shaft at each step along the shaft's profile expressed in [MPa]
--> 9) "sigma_id": a numpy array representing the ideal stress acting on the shaft at each step along the shaft's profile expressed in [MPa]

The properties can be manipulated or used via the following functions:
--> 1) "addFillet(self, radius=0, quadrant=[], zOff=0, dOff=0)": adds a fillet at position with axis offset "zOff" and radial offset "dOff" (both expressed in [mm]) with the specified radius expressed in [mm]. The "quadrant" parameter is a list specifying whether the notch is to the right of a shoulder (=[1]) or to the left of a shoulder (=[2]) or existing simply in the middle of the shaft as a groove (=[1, 2]).
--> 2) "refineProfile(self, name="", delta=0.1)": refines the profile by adding more points equally-spaced according to "delta".
--> 3) "calculateSectionProperties(self)": calculates all section properties.
--> 4) "plotProfile(self, ax=None)": plots the shaft's profile on a given axis, "ax", belonging to a matplotlib.pyplot plot.
--> 5) "calculateProfileStresses(self)": calculates the stresses at each step along the shaft's profile.
--> 6) "calculateProfileEquivalentAndIdealStress(self)": calculate equivalent and ideal stresses at each step along the shaft's profile.
--> 7) "plotStresses(self)": uses the Shaft class's "plotLoad" function to plot the stresses along the shaft's profile.
'''

import numpy as np
from math import pi
class ShaftProfile:

    # Constructor
    def __init__(self, name="", radii=np.array([]), locs=np.array([])):
        self.name = name
        # Geometry
        self.radii = np.concatenate(([0], radii, [0]))
        self.locs = np.concatenate(([locs[0]], locs, [locs[-1]]))
        # Shaft
        self.shaft = None
        # Cross-sectional properties
        self.Area = np.array([])
        self.Wb = np.array([])
        self.Wt = np.array([])
        # Stresses
        self.sigma_N = np.array([])
        self.sigma_Mb = np.array([])
        self.tau_Mt = np.array([])
        self.sigma_tot = np.array([])
        self.sigma_id = np.array([])
    
    # Overload Equal
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return all(self.radii == other.radii) and all(self.locs == other.locs)
        return False
    
    # Add Fillet
    def addFillet(self, radius=0, quadrant=[], zOff=0, dOff=0):
        for q in range(len(quadrant)):
            if quadrant[q] == 1:
                theta = np.arange(pi, 3*pi/2, 0.1)
            elif quadrant[q] == 2:
                theta = np.arange(3*pi/2, 2*pi, 0.1)
            else:
                raise ValueError("Wrong Input")
        z = radius * np.cos(theta) + zOff
        r = radius * np.sin(theta) + dOff
        before = self.locs < np.min(z)
        after = self.locs > np.max(z)
        self.radii = np.concatenate((self.radii[np.where(before)], r, self.radii[np.where(after)]))
        self.locs = np.concatenate((self.locs[np.where(before)], z, self.locs[np.where(after)]))
    
    # Refine Profile
    def refineProfile(self, name="", delta=0.1):
        if name == "":
            name = self.name
        pLen = int((self.locs[-1] - self.locs[0]) / delta + 1)
        refinedProfile = ShaftProfile(name, np.zeros(pLen), np.arange(self.locs[0], self.locs[-1] + delta / 2, delta))
        for z in range(2, len(self.locs) - 3):
            condition = np.where(np.logical_and(refinedProfile.locs >= self.locs[z], refinedProfile.locs <= self.locs[z + 1]))
            if self.locs[z] != self.locs[z + 1]:
                refinedProfile.radii[condition] = np.interp(refinedProfile.locs[condition], np.array([self.locs[z], self.locs[z+1]]), np.array([self.radii[z], self.radii[z+1]]))
        return refinedProfile
    
    # Calculate Cross-Sectional Properties
    def calculateSectionProperties(self):
        self.Area = pi * self.radii ** 2
        self.Wb = pi / 4 * self.radii ** 3
        self.Wt = pi / 2 * self.radii ** 3
    
    # Plot Profile
    def plotProfile(self, ax=None):
        ax1 = ax.twinx()
        ax1.plot(self.locs, self.radii, 'r', linewidth=1.5)
        ax1.plot(self.locs, -self.radii, 'r', linewidth=1.5)
        window = (self.locs[-1] - self.locs[0] + 20) / 2
        ax1.set_xlim(-0.1 * window, 2 * window)
        ax1.set_ylim(-window, window)
        ax1.set_ylabel("Profile [mm]")
    
    # Calculate Profile Stresses
    def calculateProfileStresses(self):
        if self.sigma_N.size != 0 and self.sigma_Mb.size != 0 and self.tau_Mt.size != 0:
            return
        sLen = len(self.locs)
        self.sigma_N = np.zeros(sLen)
        self.sigma_Mb = np.zeros(sLen)
        self.tau_Mt = np.zeros(sLen)
        self.sigma_N[np.where(self.Area != 0)] = self.shaft.N[np.where(self.Area != 0)] / self.Area[np.where(self.Area != 0)]
        self.sigma_Mb[np.where(self.Wb != 0)] = 1e3 * self.shaft.Mf[np.where(self.Wb != 0)] / self.Wb[np.where(self.Wb != 0)]
        self.tau_Mt[np.where(self.Wt != 0)] = 1e3 * self.shaft.Mt[np.where(self.Wt != 0)] / self.Wt[np.where(self.Wt != 0)]
    
    # Calculate Profile Equivalent and Ideal Stresses
    def calculateProfileEquivalentAndIdealStress(self):
        if self.sigma_tot.size != 0 and self.sigma_id.size != 0:
            return
        self.sigma_tot = self.sigma_N + self.sigma_Mb
        self.sigma_id = np.sqrt(self.sigma_tot ** 2 + 3 * self.tau_Mt ** 2)
    
    # Plot stresses
    def plotStresses(self):
        # Normal stress
        self.shaft.plotLoad(load=self.sigma_N, ylabel=r"$\sigma^{N}$ [MPa]", title=r"Normal Stress - $\sigma^{N}(z)$ [MPa]", profile=self)
        # Bending stress
        self.shaft.plotLoad(load=self.sigma_Mb, ylabel=r"$\sigma^{M_{B}}$ [MPa]", title=r"Bending Stress - $\sigma^{M_{B}}(z)$ [MPa]", profile=self)
        # Torsional stress
        self.shaft.plotLoad(load=self.tau_Mt, ylabel=r"$\tau^{M_{t}}$ [MPa]", title=r"Torsional Stress - $\tau^{M_{t}}(z)$ [MPa]", profile=self)
        # Total stress
        self.shaft.plotLoad(load=self.sigma_tot, ylabel=r"$\sigma^{tot}$ [MPa]", title=r"Resulting Normal Stress - $\sigma^{tot}(z)$ [MPa]", profile=self)
        # Equivalent stress
        self.shaft.plotLoad(load=self.sigma_id, ylabel=r"$\sigma_{id}$ [MPa]", title=r"Equivalent Stress - $\sigma_{id}(z)$ [MPa]", profile=self)


'''
This is the "Shaft" class. It inherits from "Component" class.
This class is responsible for defining the properties of a Shaft object, such as:

I) Given properties:
--> 1) "inputs": a list containing component objects representing the inputs from the shaft's perspective
--> 2) "outputs": a list containing component objects representing the outputs from the shafts's perspective
--> 3) "sups": a list containing support objects representing the bearings on the shaft
--> 4) "material": a Material object representing the material properties of the shaft
--> 5) "sections": a list of shaft section objects representing the sections at which to perform static/fatigue analysis
--> 6) "profiles": a list of shaft profile objects representing the profile of the shaft (could be more than one for different type of analyses)

II) Calculated properties:
--> 1) "N": an n-element vector representing the normal internal load distribution along the shaft's axis
--> 2) "Mx": an n-element vector representing the bending moment around x-axis  internal load distribution along the shaft's axis
--> 3) "My": an n-element vector representing the bending moment around y-axis  internal load distribution along the shaft's axis
--> 4) "Mt": an n-element vector representing the torsional moment around z-axis internal load distribution along the shaft's axis

Moreover, the user can carry out all the necessary calculations via the following functions:
--> 1) "solve(self)": implementation of virtual function from parent class resolving the external force(s)/torque(s) acting on the shaft.
--> 2) "checkTorqueEquilibrium(self)": implmenetation of virtual function from parent class to check torque equilibrium on shaft by involving external torques and forces, the latter with their corresponding locations with respect to the shaft.
--> 3) "calculateTorque(self, comp=None)": calculates the external torque acting on the shaft due to component "comp" with unknown torque. Adds the torque to the shaft's "ETs" and the component's "ETs".
--> 4) "addProfile(self, profile=None)": adds profile to the list of profiles "self.profiles" if it doesn't exist yet in the list.
--> 5) "addSection(self, sections=[])": adds the list of sections to "self.sections".
--> 6) "insertFLCF(self)": inserts fatigue limit correction factors to the sections in "self.sections".
--> 7) "calculateReactionForces(self)": calculates the forces on the bearings of the shaft by imposing equilibrium conditions with all external forces in "self.EFs".
--> 8) "performStaticVerification(self, RF=None, profile=None)": performs static verification on all sections in "self.sections" relative to the shaft's given profile.
--> 9) "performFatigueVerification(self, RF=None, profile=None)": perform fatigue verification on all sections in "self.sections" relative to the shaft's given profile.
--> 10) "calculateInternalLoads(self, RF=None, profile=None)": calculates the internal loads acting on the shaft relative to the given profile.
--> 11) "calculateStresses(self, profile=None)": calculates the stresses acting on the shaft relative the shaft's given profile.
--> 12) "calculateEquivalentAndIdealStress(self, profile=None)": calculates equivalent and ideal stresses along the shaft's given profile.
--> 13) "plotInternalLoads(self, profile=None)": plots the internal loads acting on the shaft relative to the shaft's given profile.
--> 14) "plotLoad(self, load=[], ylabel="", title="", profile=None)": generic function to plot an internal load/stress acting on the shaft relative to the shaft's given profile.
--> 15) "calculateStaticSafetyFactor(self, profile=None)": calculates the static safety factor on all sections in "self.sections" relative to the shaft's given profile.
--> 16) "calculateMeanAlternatingStress(self, profile=None)": calculates the mean and alternating stresses on all sections in "self.sections" relative the shaft's given profile.
--> 17) "calculateEquivalentStresses(self)": calculates the equivalent mean and alternating stresses on all sections in "self.sections".
--> 18) "calculateFatigueSafetyFactor(self)": calculates the fatigue safety factor on all sections in "self.sections".
'''

import numpy as np
import matplotlib.pyplot as plt
from .component import Component
from .torque import Torque
from .force import Force
from .gear import Gear
from .motor import Motor
from math import pi
class Shaft(Component):

    # Constructor
    def __init__(self, name="", inputs=[], outputs=[], axis=np.zeros(3), material=None, sups=[], loc=0):
        # Given parameters
        super().__init__(name=name, material=material, axis=axis, loc=loc, omega=inputs[0].omega)
        # Update shaft's absolute location
        if self.abs_loc.size == 0:
            self.abs_loc = -inputs[0].rel_loc + inputs[0].abs_loc
        # Inputs
        for input in inputs:
            input.onShaft = self
            input.updateLoc()
            if isinstance(input, Motor):
                input.updateForceLoc()
                input.updateTorqueLoc()
            if input.ETs.size != 0:
                self.updateETs(input.ETs)
        self.inputs = inputs
        # Outputs
        for out in outputs:
            out.onShaft = self
            out.updateLoc()
            if out.ETs.size != 0:
                self.updateETs(out.ETs)
            out.omega = self.omega
        self.outputs = outputs
        
        # Supports
        for sup in sups:
            sup.onShaft = self
            sup.updateLoc()
            sup.omega = self.omega
            sup.n = sup.omega * 30 / pi
        self.supports = sups
        # Sections
        self.sections = np.array([])
        # Profiles
        self.profiles = np.array([])
        # Internal Loads
        self.N = np.array([])
        self.Mx = np.array([])
        self.My = np.array([])
        self.Mt = np.array([])
    
    # Solve function
    def solve(self):
        unknown_Ts = 0
        unknkown_Fs = 0
        unknown_comp_T = None
        comps = np.append(self.inputs, self.outputs)
        if not self.checkTorqueEquilibrium():
            print(f"Checking solvability for {self.name}.")
            for comp in comps:
                if comp.ETs.size == 0:
                    unknown_Ts += 1
                    unknown_comp_T = comp
                else:
                    self.updateETs(comp.ETs)
                if comp.EFs.size == 0 and not isinstance(comp, Motor):
                    unknkown_Fs += 1
                else:
                    self.updateEFs(comp.EFs)
            if unknown_Ts > 1:
                print(f"{self.name}'s torque equilibrium cannot be solved.")
            elif unknown_Ts == 1:
                print(f"Solving torque equilibrium for {self.name}.")
                self.calculateTorque(unknown_comp_T)
                print(f"Torque equilibrium for {self.name} is solved.")
                if isinstance(unknown_comp_T, Gear):
                    choice = input(f"Detected torque on {unknown_comp_T.name}. Would you like to solve its torque equilibrium [y/n]: ")
                    if choice == 'y' or choice == 'Y':
                        unknown_comp_T.solve()
                        unknkown_Fs -= 1
                    else:
                        print(f"{unknown_comp_T.name}'s torque equilibrium won't be solved now.")
        else:
            print(f"No torque equilibrium to be solved for {self.name}.")
        if unknkown_Fs == 0:
            if self.checkForceEquilibrium() :
                print(f"No force equilibrium to be solved for {self.name}.")
            else:
                reaction_choice = input(f"Forces from external components on {self.name} are resolved. Do you want to calculate the reaction forces [y/n]: ")
                if reaction_choice == 'y' or reaction_choice == 'Y':
                    self.calculateReactionForces()
                    self.checkForceEquilibrium()
                else:
                    print(f"Reaction forces on {self.name} won't be calculated.")
        else:
            print(f"Force equilibrium for {self.name} cannot be solved.")
            
        
    # Check torque equilibrium
    def checkTorqueEquilibrium(self):
        print(f"Checking torque equilibrium for {self.name}.")
        valid = True
        if self.ETs.size == 0:
            valid = False
            return valid
        eq = np.zeros(3)
        eqState = False
        for ET in self.ETs:
            eq = eq + ET.torque
        if all(np.abs(eq) <= 1e-3 * np.ones(3)):
            print(f"{self.name} mainatains a torque equilibrium.")
            eqState = True
        else:
            print(f"{self.name} does not mainatain a torque equilibrium.")
        return eqState
    
    # Calculate torque
    def calculateTorque(self, comp=None):
        ET = Torque(-np.sum(self.ETs), comp.abs_loc)        
        comp.updateETs([ET])
        self.updateETs([ET])
    
    # Set shaft profile
    def addProfile(self, profile=None):
        if profile not in self.profiles:
            profile.shaft = self
            profile.calculateSectionProperties()
            self.profiles = np.append(self.profiles, profile)
    
    # Add sections
    def addSections(self, sections=[]):
        for section in sections:
            section.material = self.material
        self.sections = np.append(self.sections, sections)
    
    # Insert fatigue limit corrector factors
    def insertFLCF(self):
        for section in self.sections:
            section.addFLCF()
    
    # Calculate reaction forces
    def calculateReactionForces(self):
        # Gears axial load
        K_a = np.zeros(3)
        for EF in self.EFs:
            K_a = K_a + (EF.force * self.axis)
        # Find the bearing around which to apply moment
        index = 0
        for i in range(len(self.supports)):
            if self.supports[i].type == "Pin":
                index = i
                self.supports[index].F_tot = Force(np.zeros(3), self.supports[index].abs_loc)
        # Calculate reaction on other bearing
        for i in range(len(self.supports)):
            if i != index:
                self.supports[i].F_tot = Force(np.zeros(3), self.supports[i].abs_loc)
                supDist = self.supports[i].abs_loc - self.supports[index].abs_loc
                supDist_rec = np.array([1/d if d != 0 else 0 for d in supDist])
                for EF in self.EFs:
                    momEF = EF.moment(self.supports[index].abs_loc) #np.cross(EF.force, (EF.loc - self.supports[index].abs_loc)) * 1e-3
                    self.supports[i].F_tot.force += np.cross(momEF, supDist_rec) * 1e3
                self.updateEFs([self.supports[i].F_tot])
        # Calculate reaction around bearing with sum of external forces
        for EF in self.EFs:
            self.supports[index].F_tot.force -= EF.force
        self.updateEFs([self.supports[index].F_tot])
        # Update axial load based on configuration
        if self.supports[0].bearingType == "Tapered" and self.supports[1].bearingType == "Tapered":
            if self.supports[0].shoulder == -1:
                indA = 1
                indB = 0
                sgn = 1
            else:
                indA = 0
                indB = 1
                sgn = -1
            A_FrV = self.supports[indA].F_tot.force - self.supports[indA].F_tot.force * self.axis
            A_Fr = np.sqrt(np.sum(A_FrV * A_FrV))
            A_Y = self.supports[indA].Y
            B_FrV = self.supports[indB].F_tot.force - self.supports[indB].F_tot.force * self.axis
            B_Fr = np.sqrt(np.sum(B_FrV * B_FrV))
            B_Y = self.supports[indB].Y
            print(f"Axial reaction forces on {self.name}: ", end="")
            # Case 1
            if np.sum(K_a) > 0:
                # Factor of comparison
                fac = B_Fr / B_Y - A_Fr / A_Y
                # Case 1a
                if fac <= 0 and np.abs(np.sum(K_a)) >= 0:
                    print("Case 1a")
                    A_Fa = sgn * 0.5 * A_Fr / A_Y * self.axis
                    B_Fa = -(A_Fa + K_a)
                # Case 1b
                elif fac > 0 and np.abs(np.sum(K_a)) >= 0.5 * fac:
                    print("Case 1b")
                    A_Fa = sgn * 0.5 * A_Fr / A_Y * self.axis
                    B_Fa = -(A_Fa + K_a)
                # Case 1c
                elif fac > 0 and np.abs(np.sum(K_a)) < 0.5 * fac:
                    print("Case 1c")
                    B_Fa = -sgn * 0.5 * B_Fr / B_Y * self.axis
                    A_Fa = -(B_Fa + K_a)
            # Case 2
            else:
                # Factor of comparison
                fac = A_Fr / A_Y - B_Fr / B_Y
                # Case 2a
                if fac <= 0 and np.abs(np.sum(K_a)) >= 0:
                    print("Case 2a")
                    B_Fa = -sgn * 0.5 * B_Fr / B_Y * self.axis
                    A_Fa = -(B_Fa + K_a)
                # Case 2b
                elif fac > 0 and np.abs(np.sum(K_a)) >= 0.5 * fac:
                    print("Case 2b")
                    B_Fa = -sgn * 0.5 * B_Fr / B_Y * self.axis
                    A_Fa = -(B_Fa + K_a)
                # Case 2c
                elif fac > 0 and np.abs(np.sum(K_a)) < 0.5 * fac:
                    print("Case 2c")
                    A_Fa = sgn * 0.5 * A_Fr / A_Y * self.axis
                    B_Fa = -sgn*(A_Fa + K_a)
            self.supports[indA].F_tot.force[2] = np.sum(A_Fa)
            self.supports[indB].F_tot.force[2] = np.sum(B_Fa)
            if self.supports[0].arr == "B2B":
                self.EFs[-2].force[2] = np.sum(B_Fa)
                self.EFs[-1].force[2] = np.sum(A_Fa)
            elif self.supports[0].arr == "F2F":
                self.EFs[-2].force[2] = np.sum(A_Fa)
                self.EFs[-1].force[2] = np.sum(B_Fa)
        # Update support reaction to separate total radial force and axial force
        for support in self.supports:
            support.updateReaction()
    
    # Perform static verification
    def performStaticVerification(self, RF=None, profile=None):
        if profile == None:
            profile = self.profiles[0]
        print(f"Initiating static verification on shaft {self.name}.")
        self.calculateInternalLoads(RF=RF, profile=profile)
        plotInteralLoadsChoice = input(f"Would you like to plot shaft {self.name}'s internal loads? [y/n]: ")
        if plotInteralLoadsChoice == "Y" or plotInteralLoadsChoice == "y":
            self.plotInternalLoads(profile=profile)
        print(f"Calculating stresses on shaft {self.name} along profile '{profile.name}'.")
        self.calculateStresses(profile=profile)
        self.calculateEquivalentAndIdealStress(profile=profile)
        plotStressesChoice = input(f"Woud you like to plot profile '{profile.name}''s stresses? [y/n]: ")
        if plotStressesChoice == "Y" or plotStressesChoice == "y":
            profile.plotStresses()
        print(f"Calculating static safety factor for every user-defined section.")
        self.calculateStaticSafetyFactor(profile=profile)
        print(f"Section Name: Static Safety Factor")
        for section in self.sections:
            print(f"Section {section.name}: {section.staticSF:.2f} [-].")
    
    # Perform fatigue verification
    def performFatigueVerification(self, RF=None, profile=None):
        if profile == None:
            profile = self.profiles[0]
        print(f"Initiating fatigue verification on shaft {self.name}.")
        self.calculateInternalLoads(RF=RF, profile=profile)
        plotInteralLoadsChoice = input(f"Would you like to plot shaft {self.name}'s internal loads? [y/n]: ")
        if plotInteralLoadsChoice == "Y" or plotInteralLoadsChoice == "y":
            self.plotInternalLoads(profile=profile)
        print(f"Calculating stresses on shaft {self.name} along profile '{profile.name}'.")
        self.calculateStresses(profile=profile)
        self.calculateEquivalentAndIdealStress(profile=profile)
        plotStressesChoice = input(f"Woud you like to plot profile '{profile.name}''s stresses? [y/n]: ")
        if plotStressesChoice == "Y" or plotStressesChoice == "y":
            profile.plotStresses()
        print(f"Calculating mean and alternating stresses for every user-defined section.")
        self.calculateMeanAlternatingStress(profile=profile)
        print(f"Calculating fatigue limit corrector factors on every user-defined section.")
        self.insertFLCF()
        print(f"Calculating equivalent mean and alternating stresses along every user-defined section.")
        self.calculateEquivalentStresses()
        plotHaighDiagramChoice = input("Would you like to plot the Haigh Diagram for every defined section? [y/n]:")
        if plotHaighDiagramChoice == "Y" or plotHaighDiagramChoice == "y":
            for section in self.sections:
                section.plotHaighDiagram()
        print(f"Calculating fatigue safety factor for every user-defined section.")
        self.calculateFatigueSafetyFactor()
        print("Section Name: Fatigue Safety Factor")
        for section in self.sections:
            print(f"{section.name}: {section.fatigueSF:.2f} [-].")
        
    
    # Calculate internal loads
    def calculateInternalLoads(self, RF=None, profile=None):
        if profile == None:
            profile = self.profiles[0]
        if self.N.size != 0 and self.Mx.size != 0 and self.My.size != 0 and self.Mt.size != 0:
            print(f"Internal loads for shaft {self.name} already calculated.")
            return
        print(f"Calculating internal loads on shaft {self.name}.")
        l = len(self.profiles[0].locs)
        self.N = np.zeros(l)
        self.Mx = np.zeros(l)
        self.My = np.zeros(l)
        self.Mt = np.zeros(l)
        for EF in self.EFs:
            for i, z in enumerate(profile.locs):
                if np.dot(EF.loc - self.abs_loc, np.abs(self.axis)) <= z:
                    self.N[i] = self.N[i] - np.sum(EF.force * np.abs(self.axis))
                    mxz = EF.moment(axis=RF[1], projection=RF[2])
                    mxy = EF.moment(location=z - EF.loc, axis=RF[2], projection=RF[1])
                    self.Mx[i] = self.Mx[i] + (mxz - mxy)
                    myz = EF.moment(axis=RF[0], projection=RF[2])
                    myx = EF.moment(location=EF.loc - z, axis=RF[2], projection=RF[0])
                    self.My[i] = self.My[i] + (myz + myx)
        for ET in self.ETs:
                for i, z in enumerate(profile.locs):
                    if np.dot(ET.loc - self.abs_loc, np.abs(self.axis)) <= z:
                        self.Mt[i] = self.Mt[i] + np.sum(ET.torque)
        self.N[np.where(np.abs(self.N) < 1e-3)] = 0
        self.Mx[np.where(np.abs(self.Mx) < 1e-3)] = 0
        self.My[np.where(np.abs(self.My) < 1e-3)] = 0
        self.Mt[np.where(np.abs(self.Mt) < 1e-3)] = 0
        self.Mf = np.sqrt(self.Mx ** 2 + self.My ** 2)
    
    # Calculate stresses
    def calculateStresses(self, profile=None):
        if profile == None:
            profile = self.profiles[0]
        profile.calculateProfileStresses()
    
    # Calculate equivalent and ideal stresses
    def calculateEquivalentAndIdealStress(self, profile=None):
        if profile == None:
            profile = self.profiles[0]
        profile.calculateProfileEquivalentAndIdealStress()
    
    # Plot internal loads
    def plotInternalLoads(self, profile=None):
        if profile == None:
            profile = self.profiles[0]
        # Normal load
        self.plotLoad(load=self.N, ylabel="N [N]", title="Normal Load - N(z)", profile=profile)
        # Bending moment around x-axis
        self.plotLoad(load=self.Mx, ylabel=r"$M_{x}$ [Nm]", title=r"Bending Moment $M_{x}(z)$", profile=profile)
        # Bending moment around y-axis
        self.plotLoad(load=self.My, ylabel=r"$M_{y}$ [Nm]", title=r"Bending Moment $M_{y}(z)$", profile=profile)
        # Resulting bending moment
        self.plotLoad(load=self.Mf, ylabel=r"$M_{B}$ [Nm]", title=r"Bending Moment $M_{B}(z)$", profile=profile)
        # Torsional moment
        self.plotLoad(load=self.Mt, ylabel=r"$M_{t}$ [Nm]", title=r"Torsional Moment $M_{t}(z)$", profile=profile)
    
    # Plot load with shaft profile
    def plotLoad(self, load=[], ylabel="", title="", profile=None):
        if profile == None:
            profile = self.profiles[0]
        fig, ax = plt.subplots()
        ax.plot(profile.locs, load, 'b', linewidth = 1.5)
        ax.set_xlabel("z [mm]")
        ax.set_ylabel(ylabel)
        plt.title(title)
        if np.max(np.abs(load)) != 0:
            ax.set_ylim(-1.1 * np.max(np.abs(load)), 1.1 * np.max(np.abs(load)))
        else:
            ax.set_ylim(-1, 1)
        plt.grid()
        profile.plotProfile(ax)
        for section in self.sections:
            xs = np.ones(2) * np.sum(section.loc)
            ys = np.array([-1, 1])
            if np.max(np.abs(load)) != 0:
                ys = ys * 1.1 * np.max(np.abs(load))
            ax.plot(xs, ys, 'g--', linewidth=1.5)
            ax.text(xs[0] - 10, 0.9 * ys[0], section.name)
            ax.text(xs[0] - 10, 0.9 * ys[1], section.name)
        plt.show()
    
    # Calculate static safety factor on shaft's sections
    def calculateStaticSafetyFactor(self, profile=None):
        if profile == None:
            profile = self.profiles[0]
        for section in self.sections:
            section.calculateSectionStaticSafetyFactor(profile=profile)

    # Calculate mean and alternating stresses
    def calculateMeanAlternatingStress(self, profile=None):
        if profile == None:
            profile = self.profiles[0]
        for section in self.sections:
            section.calculateSectionMeanAlternatingStress(profile=profile)
    
    # Calculate equivalent mean and alternating stress
    def calculateEquivalentStresses(self):
        for section in self.sections:
            section.calculateSectionEquivalentStress()
    
    # Calculate fatigue safety factor on shaft's sections
    def calculateFatigueSafetyFactor(self):
        for section in self.sections:
            section.calculateSectionFatigueSafetyFactor()


'''
This is the "Motor" class. It inherits from "Component" class.
The motor class defines the properties of an electric motor based on:

I) Given properties:
--> 1) "name": a string of characters acting as a label
--> 2) "power": scalar value representing the power of the motor expressed in [W]
--> 3) "n": scalar value representing the speed expressed in [rpm]
--> 4) "torque": a 3-element vector representing the torque exerted by the motor expressed in [Nm]
 For properties (2), (3), and (4), the user can indicate either 2 of them (the 3rd will be calculated) or all of them where a check will be performed.

II) Calculated properties
--> 1) "omega": a 3-element vector representing the rotational velocity of the motor expressed in [rad/s]

The functions implemented help relocate the load(s) exerted by the motor once placed on a shaft.
'''
import numpy as np
from .component import Component
from .force import Force
from .torque import Torque
from math import pi

class Motor(Component):
    
    # Constructor
    def __init__(self, name="", loc=0, power=0, n=0, torque=Torque(np.array([]), np.array([])), axis=np.array([0, 0, 0])):
        # Given properties
        super().__init__(name=name, material=None, axis=axis, loc=loc)
        # Check for valid input
        if power == 0 and n == 0 and torque.torque.size == 0:
            raise ValueError("A minimum of 2 out of 3 inputs are necessary: power, n, and torque.")
        elif power != 0 and n != 0 and torque.torque.size != 0:
            omega = n * pi / 30 * self.axis
            omega_mag = np.sqrt(np.sum(omega * omega))
            if power != torque.mag() * omega_mag:
                raise ValueError(f"Provided given is incoherent: {power} != {torque.mag() * omega_mag}.")
        # Check which inputs are given
        if power == 0:
            omega = n * pi / 30 * self.axis
            omega_mag = np.sqrt(np.sum(omega * omega))
            power = torque.mag() * omega_mag
        elif n == 0:
            omega = power / torque.mag() * self.axis
            omega_mag = np.sqrt(np.sum(omega * omega))
            n = omega_mag * 30 / pi
        elif torque.torque.size == 0:
            omega = n * pi / 30
            torque_mag = power / omega
            torque = torque_mag * self.axis

        self.power = power
        self.n = n
        self.omega = omega
        if self.abs_loc.size != 0:
            location = self.abs_loc
        else:
            location = self.rel_loc
        self.updateETs([Torque(torque=torque, loc=location)])
        self.updateEFs([Force(force=np.array([0, 0, 0]), loc=location)])
    
    # Update Force Location
    def updateForceLoc(self):
        for EF in self.EFs:
            EF.loc = self.abs_loc
    
    # Update Torque Location
    def updateTorqueLoc(self):
        for ET in self.ETs:
            ET.loc = self.abs_loc


'''
This is the "Gear" class. It inherits from "Component" class.
This class is responsible for defining the properties of a Gear object (spur/helical), such as:

I) Given Properties
--> 1) "name": a string of characters acting as a label
--> 2) "m_n": normal module of the gear expressed in [mm]
--> 3) "z": number of teeth of the gear
--> 4) "psi": helix angle expressed in [rad]
--> 5) "phi_n": normal pressure angle expressed in [rad]
--> 6) "Q_v": transmission accuracy
--> 7) "FW": tooth face width expressed in [mm]
--> 8) "material": a Material object representing the material properties of the gear

II) Calculated parameters
--> 1) "p_n": normal circular pitch expressed in [mm]
--> 2) "p_t": transverse circular pitch expressed in [mm]
--> 3) "p_x": axial pitch expressed in [mm]
--> 4) "m_t": transverse module expressed in [mm]
--> 5) "d": pitch diameter expressed in [mm]
--> 6) "phi_t": transverse pressure angle expressed in [rad]
--> 7) "z_p": virtual number of teeth
--> 8) "phi_b": base helix angle expressed in [rad]
--> 9) "h_a": addendum height expressed in [mm]
--> 10) "h_f": dedendum height expressed in [mm]
--> 11) "h": tooth height expressed in [mm]
--> 12) "d_a": addendum diameter expressed in [mm]
--> 13) "d_f": dedendum diameter expressed in [mm]

III) Gear tooth verification parameters
--> 1) "sigma_max_fatigue": maximum gear tooth bending stress for fatigue expressed in [MPa]
--> 2) "power_source": a string of characters deining the power source type
--> 3) "K_0": overload factor
--> 4) "t_R": difference between dedendum radius and internal gear radius expressed in [mm]
--> 5) "m_B": ratio between t_R and tooth height
--> 6) "K_B": rim-thickness factor
--> 7) "K_v": dynamic factor
--> 8) "C_mc": coefficient 1 for load distribution factor
--> 9) "C_pf": coefficient 2 for load distribution factor
--> 10) "C_pm": coefficient 3 for load distribution factor
--> 11) "C_ma": coefficient 4 for load distribution factor
--> 12) "C_e":coefficient 5 for load distribution factor
--> 13) "Y": coefficient 6 for load distribution factor
--> 14) "K_H": load distribution factor
--> 15) "K_S": size factor
--> 16) "J_p": geometry factor
--> 17) "J_pp": modifying factor
--> 18) "Y_J": bending strength geometry factor
--> 19) "sigma_FP": bending fatigue strength expressed in [MPa]
--> 20) "Y_N": stress cycle life factor (bending)
--> 21) "Y_theta": temperature factor
--> 22) "Y_Z": reliability factor
--> 23) "bendingSF": bending safety factor
--> 24) "sigma_max_pitting": maximum gear contact (pitting resistance) stess expressed in [MPa]
--> 25) "Z_E": elastic coefficient expressed in [MPa^(1/2)]
--> 26) "Z_I": surface strength geometry
--> 27) "m_N": load sharing ratio
--> 28) "sigma_HP": contact strength expressed in [MPa]
--> 29) "Z_N": stress cycle life factor (wear)
--> 30) "Z_W": hardness-ratio factor
--> 31) "wearSF": wear safety factor

Moreover, the user can carry out all the necessary calculations via the following functions:
--> 1) "solve(self)": implementation of virtual function from parent class resolving the external force(s)/torque(s) acting on the gear.
--> 2) "checkTorqueEquilibrium(self)": implmenetation of virtual function from parent class to check torque equilibrium on gear by involving external torques and forces, the latter with their corresponding locations with respect to the gear's location.
--> 3) "calculateTorque(self)": calculates the total external torque on the gear based on sum of moments of all external forces and adds it to "self.ETs".
--> 4) "calculateForces(self, mesh=None)": calculates the force acting on the gear due to the specific given "mesh" and adds it to "self.EFs". Updates the mesh forces as well.
--> 5) "analyseGearToothBending(self, mesh=None, powerSource="", drivenMachine="", dShaft=0, Ce=0, teethCond="", lShaft=0, useCond="", sigma_FP=0, b_YN=0, e_YN=0, N=0, temp=0, rel=0)": performs gear tooth bending analysis and produces the maximum gear tooth bending stress express in [MPa] and the gear tooth bending safety factor.
--> 6) "calculateSigmaMaxFatigue(self, mesh=None, powerSource="", drivenMachine="", dShaft=0, Ce=0, teethCond="", lShaft=0, useCond="")": calculates the maximum gear tooth bending stress expressed in [MPa].
--> 7) "calculateBendingSF(self, sigma_FP=0, b_YN=0, e_YN=0, N=0, temp=0, rel=0)": calculates the gear tooth bending safety factor.
--> 8) "analyseGearToothPitting(self, mesh=None, Z_R=0, sigma_HP=0, b_ZN=0, e_ZN=0, N=0)": perform gear tooth pitting analysis and produces the maximum gear tooth contact stress expressed in [MPa] and the wear safety factor.
--> 9) "calculateSigmaMaxPitting(self, mesh=None, Z_R=0)": calculates the maximum gear tooth contact stress expressed in [MPa].
--> 10) "calculateWearSF(self, sigma_HP=0, b_ZN=0, e_ZN=0, N=0, mesh=None)": calculates the wear safety factor.
'''
import numpy as np
from .component import Component
from .force import Force
from .torque import Torque
from math import pi, cos, sin, tan, atan, ceil, sqrt, log
from .makima2dInterpolator import Makima2DInterpolator
class Gear(Component):

    # Overload factor reference matrix
    K0_ref = np.array([[1, 1.25, 1.75], [1.25, 1.5, 2], [1.5, 1.75, 2.25]])
    # Size factor reference graph
    z_ref = np.concatenate((np.arange(12, 23, 1), np.arange(24, 31, 2), np.array([34, 38, 43, 50, 60, 75, 100, 150, 300, 400])))
    Y_ref = np.array([0.245, 0.261, 0.277, 0.29, 0.296, 0.303, 0.309, 0.314, 0.322, 0.328, 0.331, 0.337, 0.346, 0.353, 0.359, 0.371, 0.384, 0.397, 0.409, 0.422, 0.435, 0.447, 0.46, 0.472, 0.48])
    # Bending strength geometry factor reference graphs
    psi_Jp_ref = np.arange(5, 35, 1)
    z_Jp_ref = np.array([20, 30, 60, 150, 500])
    Jp_ref = np.array([[0.465, 0.475, 0.48, 0.487, 0.492, 0.495, 0.497, 0.5, 0.502, 0.505, 0.506, 0.507, 0.508, 0.507, 0.506, 0.505, 0.502, 0.5, 0.497, 0.494, 0.49, 0.486, 0.48, 0.476, 0.471, 0.465, 0.458, 0.452, 0.445, 0.439], 
                       [0.525, 0.53, 0.535, 0.54, 0.542, 0.547, 0.55, 0.552, 0.553, 0.554, 0.555, 0.556, 0.555, 0.554, 0.552, 0.551, 0.549, 0.545, 0.54, 0.537, 0.532, 0.527, 0.52, 0.515, 0.507, 0.5, 0.492, 0.484, 0.476, 0.47], 
                       [0.58, 0.585, 0.595, 0.6, 0.602, 0.605, 0.61, 0.612, 0.614, 0.615, 0.615, 0.615, 0.614, 0.611, 0.608, 0.605, 0.6, 0.595, 0.59, 0.582, 0.577, 0.571, 0.56, 0.555, 0.545, 0.537, 0.527, 0.517, 0.507, 0.495], 
                       [0.62, 0.63, 0.635, 0.64, 0.645, 0.647, 0.652, 0.655, 0.657, 0.657, 0.656, 0.655, 0.654, 0.65, 0.645, 0.64, 0.635, 0.63, 0.622, 0.617, 0.61, 0.6, 0.592, 0.582, 0.575, 0.56, 0.552, 0.541, 0.53, 0.518], 
                       [0.65, 0.655, 0.66, 0.665, 0.67, 0.675, 0.677, 0.68, 0.681, 0.681, 0.681, 0.681, 0.679, 0.676, 0.673, 0.667, 0.66, 0.655, 0.647, 0.64, 0.631, 0.62, 0.613, 0.602, 0.592, 0.58, 0.57, 0.557, 0.545, 0.532]])
    psi_Jpp_ref = np.arange(5, 36, 5)
    z_Jpp_ref = np.array([20, 30, 50, 75, 150, 500])
    Jpp_ref = np.array([[0.927, 0.929, 0.93, 0.932, 0.938, 0.943, 0.953],
                        [0.952, 0.954, 0.957, 0.959, 0.961, 0.965, 0.973],
                        [0.98, 0.981, 0.981, 0.982, 0.983, 0.987, 0.992],
                        np.ones(7),
                        [1.02, 1.02, 1.02, 1.019, 1.017, 1.013, 1.01],
                        [1.038, 1.037, 1.035, 1.032, 1.03, 1.026, 1.02]])
    rel_ref = np.array([.9999, .999, .99, .9, .5])
    YZ_ref = np.array([1.5, 1.25, 1, 0.85, 0.7])
    ZE_ref = np.array([[191, 181, 179, 174, 162, 158], 
                       [181, 174, 172, 168, 158, 154], 
                       [179, 172, 170, 166, 156, 152], 
                       [174, 168, 166, 163, 154, 149], 
                       [162, 158, 156, 154, 145, 141], 
                       [158, 154, 152, 149, 141, 137]])

    # Constructor
    def __init__(self, name="", axis=np.zeros(3), loc=0, m_n=0, z=0, psi=0, phi_n=0, Q_v=0, FW=0, material=None):
        # Given parameters
        super().__init__(name=name, material=material, axis=axis, loc=loc)
        self.m_n = m_n
        self.z = z
        self.psi = psi * pi / 180
        self.phi_n = phi_n * pi / 180
        self.Q_v = Q_v
        self.FW = FW
        # Calculated parameters
        self.p_n = self.m_n * pi
        self.p_t = self.p_n / cos(self.psi)
        self.p_x = self.p_t / tan(self.psi)
        self.m_t = self.m_n / cos(self.psi)
        self.d = self.m_t * self.z
        self.phi_t = atan(tan(self.phi_n) / cos(self.psi))
        self.z_p = ceil(self.z / (cos(self.psi) ** 3))
        self.phi_b = atan(tan(self.psi) * cos(self.phi_n))
        self.h_a = self.m_n
        self.h_f = 1.25 * self.m_n # assumption
        self.h = self.h_a + self.h_f
        self.d_a = self.d + 2 * self.h_a
        self.d_f = self.d - 2 * self.h_f
        self.meshes = np.array([])
    
    # Solve function
    def solve(self):
        if self.meshes.size == 0:
            print(f"No meshes on gear {self.name} to solve. Please define configuration properly.")
            return
        unknown_Ts = 0
        unknown_Fs = 0
        unknown_mesh = None
        if not self.checkTorqueEquilibrium():
            print(f"Checking solvability for {self.name}.")
            if self.ETs.size == 0:
                unknown_Ts += 1
            for mesh in self.meshes:
                if all(mesh.F.force == np.zeros(3)):
                    unknown_Fs += 1
                    unknown_mesh = mesh
                else:
                    if self.name == mesh.drivingGear.name:
                       sign = -1
                    else:
                        sign = 1
                    self.updateEFs([Force(sign * mesh.F.force, mesh.F.loc)])
            if unknown_Fs + unknown_Ts > 1:
                print(f"{self.name}'s equilibrium cannot be solved.")
                return
            elif unknown_Ts == 1:
                print(f"Solving torque equilibrium for {self.name}.")
                self.calculateTorque()
                self.onShaft.updateETs(self.ETs)
            elif unknown_Fs == 1:
                print(f"Solving forces on {self.name} due to mesh {unknown_mesh.name}.")
                self.calculateForces(unknown_mesh)
                self.onShaft.updateEFs(self.EFs)  
            self.checkTorqueEquilibrium()              
        else:
            print(f"Nothing to be solved for {self.name}.")
    
    # Check torque equilibrium
    def checkTorqueEquilibrium(self):
        print(f"Checking torque equilibrium for {self.name}.")
        valid = False
        if self.EFs.size != 0 or self.ETs.size != 0:
            valid = True
        else:
            return valid
        eq = np.zeros(3)
        eqState = False

        for ET in self.ETs:
            eq += ET.torque
        for EF in self.EFs:
            eq += EF.moment(self.abs_loc, self.axis) #np.cross(EF.force, (EF.loc - self.abs_loc) * 1e-3) * np.abs(self.axis)
        if all(np.abs(eq) <= 1e-3 * np.ones(3)):
            print(f"{self.name} mainatains a torque equilibrium.")
            eqState = True
        else:
            print(f"{self.name} does not mainatain a torque equilibrium.")
        return eqState

    # Calculate Torque
    def calculateTorque(self):
        ET = Torque(np.zeros(3), self.abs_loc)
        for EF in self.EFs:
            ET.torque -= EF.moment(self.abs_loc, self.axis)
        self.updateETs([ET])

    # Calculate Forces
    def calculateForces(self, mesh=None):
        ET = self.ETs[0].torque
        for EF in self.EFs:
            ET -= EF.moment(self.abs_loc, self.axis)
        sign = 1
        if self.name == mesh.drivingGear.name:
            sign = -1
        F_t = np.cross(ET, 2 / self.d * mesh.radiality) * 1e3
        mesh.F_t.force = sign * F_t
        magF_t = sqrt(np.sum(F_t * F_t))
        F_r = sign * magF_t * tan(self.phi_n) / cos(self.psi) * mesh.radiality
        mesh.F_r.force = sign * F_r
        F_a = np.sign(np.sum(self.onShaft.axis)) * np.sign(self.psi) * np.abs(np.cross(mesh.radiality, F_t * tan(np.abs(self.psi))))
        mesh.F_a.force = -F_a
        self.updateEFs([Force(F_t + F_r + F_a, mesh.loc)])
        mesh.F.force = mesh.F_t.force + mesh.F_r.force + mesh.F_a.force

    # Perform gear tooth bending analysis
    def analyseGearToothBending(self, mesh=None, powerSource="", drivenMachine="", dShaft=0, Ce=0, teethCond="", lShaft=0, useCond="",
                                sigma_FP=0, b_YN=0, e_YN=0, N=0, temp=0, rel=0):
        print(f"Initiating gear tooth bending analysis for gear {self.name}.")
        print(f"Calculating maximum gear tooth bending stress for fatigue.")
        self.calculateSigmaMaxFatigue(mesh=mesh, powerSource=powerSource, drivenMachine=drivenMachine, dShaft=dShaft,
                                      Ce=Ce, teethCond=teethCond, lShaft=lShaft, useCond=useCond)
        print(f"Maximum gear tooth bending stress for fatigue for gear {self.name}: {self.sigma_max_fatigue:.2f} [MPa].")
        print(f"Calculating bending safety factor for gear {self.name}.")
        self.calculateBendingSF(sigma_FP=sigma_FP, b_YN=b_YN, e_YN=e_YN, N=N, temp=temp, rel=rel)
        print(f"Gear tooth bending safety factor for gear {self.name}: {self.bendingSF:.2f} [-].")
    
    # Maximum tooth gear bending stress equation for fatigue
    def calculateSigmaMaxFatigue(self, mesh=None, powerSource="", drivenMachine="", dShaft=0, Ce=0, teethCond="", lShaft=0, useCond=""):
        fw_min = min(mesh.drivingGear.FW, mesh.drivenGear.FW)
        # Overload factor K_0
        if powerSource == "Uniform":
            i = 1
        elif powerSource == "Light shock":
            i = 2
        elif powerSource == "Medium shock":
            i = 3
        else:
            raise ValueError("Wrong input for power source")
        if drivenMachine == "Uniform":
            j = 1
        elif drivenMachine == "Moderate shock":
            j = 2
        elif drivenMachine == "Heavy shock":
            j = 3
        else:
            raise ValueError("Wrong input for driven machine")
        self.K_0 = self.__class__.K0_ref[i-1][j-1]
        # Rim-thickness factor
        self.t_R = (self.d_f - dShaft) / 2
        self.m_B = self.t_R / self.h
        if self.m_B < 1.2:
            self.K_B = 1.6 * log(2.242 / self.m_B)
        else:
            self.K_B = 1
        # Dynamic factor
        v = np.sum(np.abs(self.omega)) * self.d * 1e-3
        B = 0.25 * (12 - self.Q_v) ** (2/3)
        A = 50 + 56 * (1 - B)
        self.K_v = ((A + sqrt(200 * v)) / A) ** B
        # Load distribution factor
        self.C_e = Ce
        if teethCond == "uncrowned teeth":
            self.C_mc = 1
        elif teethCond == "crowned teeth":
            self.C_mc = 0.8
        else:
            raise ValueError("Invalid input for teeth condition.")
        if fw_min / 25.4 <= 1:
            self.C_pf = fw_min / 10 / self.d - 0.025
        elif fw_min / 25.4 > 1 and fw_min / 25.4 <= 17:
            self.C_pf = fw_min / 10 / self.d - 0.0375 + 0.0125 * fw_min / 25.4
        else:
            self.C_pf = fw_min / 10 / self.d - 0.1109 + 0.0207 * fw_min / 25.4 - 0.000228 * (fw_min / 25.4) ** 2
        S1 = abs(lShaft / 2 - np.sum(np.abs(self.rel_loc)))
        if S1 / lShaft < 0.175:
            self.C_pm = 1
        else:
            self.C_pm = 1.1
        if useCond == "Open gearing":
            A = 0.247
            B = .0167
            C = -0.765 * 1e-4
        elif useCond == "Commercial, enclosed units":
            A = 0.127
            B = 0.0158
            C = -0.93 * 1e-4
        elif useCond == "Precision, enclosed units":
            A = 0.0675
            B = 0.0128
            C = -0.926 * 1e-4
        elif useCond == "Extraprecision enclosed gear units":
            A = 0.0036
            B = 0.0102
            C = -0.822 * 1e-4
        else:
            raise ValueError("wrong input for use.")
        self.C_ma = A + B * fw_min / 25.4 + C * (fw_min / 25.4) ** 2
        self.K_H = 1 + self.C_mc * (self.C_pf * self.C_pm + self.C_ma * self.C_e)
        # Size factor
        self.Y = np.interp(self.z_p, self.z_ref, self.Y_ref)
        self.K_S = 0.843 * (fw_min * self.m_t * sqrt(self.Y)) ** 0.0535
        # Bending strength geometry factor
        interp_func1 = Makima2DInterpolator(self.psi_Jp_ref, self.z_Jp_ref, self.Jp_ref)
        self.J_p = interp_func1(abs(self.psi * 180 / pi), self.z)
        interp_func2 = Makima2DInterpolator(self.psi_Jpp_ref, self.z_Jpp_ref, self.Jpp_ref)
        self.J_pp = interp_func2(abs(self.psi * 180 / pi), self.z)
        self.Y_J = self.J_p * self.J_pp
        self.sigma_max_fatigue = np.sum(np.abs(mesh.F_t.force)) * self.K_0 * self.K_B * self.K_v * self.K_H * self.K_S / fw_min / self.m_t / self.Y_J
    
    # Calulcate bending safety factor
    def calculateBendingSF(self, sigma_FP=0, b_YN=0, e_YN=0, N=0, temp=0, rel=0):
        self.sigma_FP = sigma_FP
        self.Y_N = b_YN * N ** e_YN
        if temp <= 120:
            self.Y_theta = 1
        if rel in self.__class__.rel_ref:            
            ind = np.where(self.__class__.rel_ref == rel)
            self.Y_Z = self.__class__.YZ_ref[ind][0]
        else:
            self.Y_Z = np.interp(rel, self.__class__.rel_ref, self.__class__.YZ_ref)[0]
        self.bendingSF = self.sigma_FP * self.Y_N / self.sigma_max_fatigue / self.Y_theta / self.Y_Z

    # Perform gear tooth pitting analysis
    def analyseGearToothPitting(self, mesh=None, Z_R=0, sigma_HP=0, b_ZN=0, e_ZN=0, N=0):
        print(f"Initiating gear tooth pitting analysis for gear {self.name}.")
        print(f"Calculating maximum gear contact stress.")
        self.calculateSigmaMaxPitting(mesh=mesh, Z_R=Z_R)
        print(f"Maximum gear contact stress for gear {self.name}: {self.sigma_max_pitting:.2f} [MPa].")
        print(f"Calculating wear safety fator for gear {self.name}.")
        self.calculateWearSF(sigma_HP=sigma_HP, b_ZN=b_ZN, e_ZN=e_ZN, N=N, mesh=mesh)
        print(f"Wear safety factor for gear {self.name}: {self.wearSF:.2f} [-].")
    
    # Calculate sigma max pitting
    def calculateSigmaMaxPitting(self, mesh=None, Z_R=0):
        # Elastic coefficient
        if mesh.drivingGear.material.name == "Steel":
            i = 1
        elif mesh.drivingGear.material.name == "Malleable iron":
            i = 2
        elif mesh.drivingGear.material.name == "Nodular iron":
            i = 3
        elif mesh.drivingGear.material.name == "Cast iron":
            i = 4
        elif mesh.drivingGear.material.name == "Aluminum bronze":
            i = 5
        elif mesh.drivingGear.material.name == "Tin bronze":
            i = 6
        else:
            raise ValueError("Wrong input for pinion material.")
        if mesh.drivenGear.material.name == "Steel":
            j = 1
        elif mesh.drivenGear.material.name == "Malleable iron":
            j = 2
        elif mesh.drivenGear.material.name == "Nodular iron":
            j = 3
        elif mesh.drivenGear.material.name == "Cast iron":
            j = 4
        elif mesh.drivenGear.material.name == "Aluminum bronze":
            j = 5
        elif mesh.drivenGear.material.name == "Tin bronze":
            j = 6
        else:
            raise ValueError("Wrong input for gear material.")
        self.Z_E = self.__class__.ZE_ref[i-1][j-1]
        # Surface strength geometry
        pG = mesh.drivingGear
        gG = mesh.drivenGear
        f1_ZA = sqrt((pG.d / 2) ** 2 + pG.h_a ** 2)
        f2_Z = (pG.d + gG.d) * sin(self.phi_t) / 2
        Z_A = min([f1_ZA, f2_Z])
        f1_ZB = sqrt((gG.d / 2) ** 2 + gG.h_a ** 2)
        Z_B = min([f1_ZB, f2_Z])
        Z = Z_A + Z_B - f2_Z
        self.m_N = self.p_n * cos(self.phi_n) / 0.95 / Z
        if mesh.type == "External":
            self.Z_I = cos(self.phi_t) * sin(self.phi_t) * mesh.m_G / 2 / self.m_N / (mesh.m_G + 1)
        elif mesh.type == "Internal":
            self.Z_I = cos(self.phi_t) * sin(self.phi_t) * mesh.m_G / 2 / self.m_N / (mesh.m_G - 1)
        else:
            raise ValueError("Gear mesh type invalid.")
        self.sigma_max_pitting = self.Z_E * np.sqrt(np.sum(np.abs(mesh.F_t.force)) * self.K_0 * self.K_v * self.K_S * self.K_H * Z_R / self.FW / self.d / self.Z_I)
    
    # Calculate wear safety factor
    def calculateWearSF(self, sigma_HP=0, b_ZN=0, e_ZN=0, N=0, mesh=None):
        self.sigma_HP = sigma_HP
        self.Z_N = b_ZN * N ** e_ZN
        HB_ratio = mesh.drivingGear.material.HB / mesh.drivenGear.material.HB
        if HB_ratio < 1.2:
            A_p = 0
        elif HB_ratio >= 1.2 and HB_ratio <= 1.7:
            A_p = (8.98 * HB_ratio - 8.29) * 1e-3
        else:
            A_p = 0.00698
        self.Z_W = 1 + A_p * (mesh.m_G - 1)
        self.wearSF = self.sigma_HP * self.Z_N * self.Z_W / self.sigma_max_pitting / self.Y_theta / self.Y_Z


'''
This is the "Oil" class.
It is used to define the oil properties of the oil being used.

The properties are:
--> 1) "name": a string of characters acting as a label
--> 2) "temperature": the oil operating temperature expressed in [degC]
--> 3) "v1": Rated viscosity of the oil expressed in [mm^2/s]
--> 4) "v": viscosity of the oil expressed in [mm^2/s]
--> 5) "k": viscosity ratio of the oil
'''
class Oil:

    # Constructor
    def __init__(self, name="", temp=0, v1=0, v=0):
        self.name = name
        self.temperature = temp
        self.v1 = v1
        self.v = v
        self.k = self.v / self.v1


'''
This is the "Force" class.

It defines a force vector based on two simple properties:
--> 1) "force": a 3-element force vector representing the force expressed in [N]
--> 2) "loc": a scalar or 3-element vector representing the point of application of the force expressed in [mm]
'''
import numpy as np
class Force:

    # Constructor
    def __init__(self, force=np.zeros(3), loc=0):
        self.force = force
        self.loc = loc
    
    # Overload Addition
    def __add__(obj1, obj2):
        return obj1.force + obj2.force
    
    # Overload Subtraction
    def __sub__(obj1, obj2):
        return obj1.force - obj2.force
    
    # Overload Negative
    def __neg__(obj):
        return -obj.force
    
    # Overload Equal
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return all(self.force == other.force) and all(self.loc == other.loc)
        return False
    
    # Overload Call
    def __call__(self):
        print(f"Force: {self.force}\nLoc: {self.loc}")

    # Calculate Magnitude
    def mag(self):
        return np.sqrt(np.sum(self.force * self.force))
    
    # Calculate moment at location around axis
    def moment(self, location=np.array([]), axis=np.array([]), projection=np.array([])):
        if projection.size == 0:
            if axis.size == 0:
                return np.cross(self.force, (self.loc - location) * 1e-3)
            else:
                return np.cross(self.force, (self.loc - location) * 1e-3) * np.abs(axis)
        else:
            if location.size == 0:
                return np.sum(np.cross(self.force * projection, self.loc * axis * 1e-3))
            else:
                return np.sum(np.cross(self.force * projection, location * axis * 1e-3))


'''
This is the "Shaft Section" class.
It's used to define certain sections of a "Shaft" component along its profile.

It has the following properties:
I) Given Properties:
--> 1) "name": a string of characters acting as a label
--> 2) "loc": a scalar/3-D element vector representing the location of the section along the shaft
--> 3) "d": a scalar representing the diameter of the shaft at the specified cross-section expressed in [mm]
--> 4) "Ra": a scalar representing the surface roughness at the specified cross-section expressed in [micrometers]
--> 5) "material": a Material object representing the material of the shaft

II) Calculated Properties:
--> 1) "sigma_m_Mb": mean bending stress [MPa]
--> 2) "sigma_a_Mb": alternating bending stress [MPa]
--> 3) "sigma_m_N": mean normal stress [MPa]
--> 4) "sigma_a_N": alternating normal stress [MPa]
--> 5) "tau_m_Mt": mean torsional stress [MPa]
--> 6) "tau_a_Mt": alternating torsional stress [MPa]
--> 7) "Kt_B": bending stress concentration rasiser
--> 8) "Kt_N": normal stress concentration raiser
--> 9) "Kt_T": torsion stress concentration raiser
--> 10) "Kf_B": bending stress concentration factor
--> 11) "Kf_N": normal stress concentration factor
--> 12) "Kf_T": torsion stress concentration factor

The properties can be manipulated or used via the following functions:
--> 1) "appendKf(self, Kf=0, loadType="")": appends a stress concentration factor according to the given load type.
--> 2) "calculateFatigueIntensificationFactor(self)": calculates the fatigue intensification factor for every load type and appends it to the list of concentration factors.
--> 3) "addNotchSensitivity(self, notchRadius=0, sigma_u=0)": calculates the notch sensitivity and then calls the "calculateFatigueIntensificatinoFactor" function if geometric stress raisers are not "None".
--> 4) "addGeometricStressRaiser(self, r2d=0, D2d=0)": calculates geometric stress raisers then calls the "calculateFatigueIntensificationFactor" function.
--> 5) "addFLCF(self)": calculates fatigue limit corrector factors then calls the material's "calculateComponentFatigueLimit" to calculate the shaft's specific fatigue limit at section object.
--> 6) "calculateSectionEquivalentStress(self)": calculates section's equivalent mean and laternating stresses according to the Shighley method.
--> 7) "plotHaighDiagram(self)": plots the Haigh Diagram representing the operating point on section object.
--> 8) "calculateSectionStaticSafetyFactor(self, profile=None)": calculates the section's safety factor relative to the shaft's given profile.
--> 9) "calculateSectionMeanAlternatingStress(self, profile=None)": calculates the section's mean and alternating stresses relative the shaft's given profile.
--> 10) "calculateSectionFatigueSafetyFactor(self)": calculates the section's fatigue safety factor.
'''
from .notchSensitivity import NotchSensitivity
from .fatigueLimitCorrectorFactors import FatigueLimitCorrectorFactors
from .geometricStressRaiser import GeometricStressRaiser
from math import sqrt
import numpy as np
import matplotlib.pyplot as plt
class ShaftSection:

    # Constructor
    def __init__(self, name="", loc=0, d=0, Ra=0, material=None):
        self.name = name
        self.loc = loc
        self.d = d
        self.Ra = Ra
        self.material = material
        # initialize mean and alternating stresses to 0
        self.sigma_m_Mb = 0
        self.sigma_a_Mb = 0
        self.sigma_m_N = 0
        self.sigma_a_N = 0
        self.tau_m_Mt = 0
        self.tau_a_Mt = 0
        # initialize stress concentration raisers
        self.Kt_B = 0
        self.Kt_N = 0
        self.Kt_T = 0
        # initialize stress concentration factor
        self.Kf_N = 1
        self.Kf_B = 1
        self.Kf_T = 1
        # initialize static safety factor
        self.staticSF = 0
    
    # Append fatigue stress intensification factor
    def appendKf(self, Kf=0, loadType=""):
        for kf, lt in zip(Kf, loadType):
            if lt == "Normal":
                self.Kf_N = self.Kf_N * kf
            elif lt == "Bending":
                self.Kf_B = self.Kf_B * kf
            elif lt == "Torsion":
                self.Kf_T = self.Kf_T * kf
            else:
                raise ValueError(f"'{lt}' is Invalid")
            
    # Calculate fatigue stress intensification factor
    def calculateFatigueIntensificationFactor(self):
        self.appendKf([1 + self.q.qReq * (self.Kt_B - 1)], ["Bending"])
        self.appendKf([1 + self.q.qReq * (self.Kt_N - 1)], ["Normal"])
        self.appendKf([1 + self.q.qReq * (self.Kt_T - 1)], ["Torsion"])

    # Notch Sensitivity
    def addNotchSensitivity(self, notchRadius=0, sigma_u=0):
        self.q = NotchSensitivity(notchRadius=notchRadius, sigma_u=sigma_u)
        if self.Kt_B == None and self.Kt_N == None and self.Kt_T == None:
            self.calculateFatigueIntensificationFactor(self)
    
    #Geomtric Stress Raiser
    def addGeometricStressRaiser(self, r2d=0, D2d=0):
        temp = GeometricStressRaiser(r2d, D2d)
        self.Kt_B = temp.Kt_Breq
        self.Kt_N = temp.Kt_Nreq
        self.Kt_T = temp.Kt_Treq
        self.calculateFatigueIntensificationFactor()
    
    # Add fatigue limit corrector factors
    def addFLCF(self):
        self.FLCF = FatigueLimitCorrectorFactors(self)
        self.material.calculateComponentFatigueLimit(self)
    
    # Calculate Equivalent Stress
    def calculateSectionEquivalentStress(self):
        self.sigma_a_eq = sqrt((self.Kf_B * self.sigma_a_Mb + self.Kf_N * self.sigma_a_N/0.85) ** 2 + 3 * (self.Kf_T * self.tau_a_Mt) ** 2)
        self.sigma_m_eq = sqrt((self.Kf_B * self.sigma_m_Mb + self.Kf_N * self.sigma_m_N/0.85) ** 2 + 3 * (self.Kf_T * self.tau_m_Mt) ** 2)

    # Plot Haigh Diagram
    def plotHaighDiagram(self):
        # Calculate coordinates of specific points
        coeff1 = np.polyfit(np.array([0, self.material.sigma_y]), np.array([self.material.sigma_y, 0]), deg=1)
        coeff2 = np.polyfit(np.array([0, self.material.sigma_u]), np.array([self.material.sigma_Dm1C, 0]), deg=1)
        inter_x = (coeff2[1] - coeff1[1]) / (coeff1[0] - coeff2[0])
        inter_y = np.polyval(coeff1, inter_x)
        coeff3 = np.polyfit(np.array([0, self.sigma_m_eq]), np.array([0, self.sigma_a_eq]), deg=1)
        P_x = (coeff2[1] - coeff3[1]) / (coeff3[0] - coeff2[0])
        if P_x > inter_x:
            P_x = (coeff1[1] - coeff3[1]) / (coeff3[0] - coeff1[0])
        P_y = np.polyval(coeff3, P_x)
        # Plot
        figure, ax = plt.subplots()
        ax.plot(np.array([0, inter_x, self.material.sigma_y]), np.array([self.material.sigma_Dm1C, inter_y, 0]), 'k', linewidth=1.5)
        ax.plot(self.sigma_m_eq, self.sigma_a_eq, 'm*', linewidth=1.5)
        ax.plot(np.array([0, inter_x, self.material.sigma_u]), np.array([self.material.sigma_y, inter_y, 0]), 'k--', linewidth=1.5)
        ax.plot(0, self.material.sigma_Dm1C, "bo", linewidth=1.5)
        ax.plot(0, self.material.sigma_y, "bo", linewidth=1.5)
        ax.plot(self.material.sigma_y, 0, "bo", linewidth=1.5)
        ax.plot(self.material.sigma_u, 0, "bo", linewidth=1.5)
        ax.plot(0, P_y, "bo", linewidth=1.5)
        ax.plot(np.array([0, P_x]), np.array([0, P_y]), "g--", linewidth=1.5)
        ax.plot(np.array([0, P_x]), np.array([P_y, P_y]), "g--", linewidth=1.5)
        ax.text(self.sigma_m_eq, self.sigma_a_eq - 35, "P" + self.name[1])
        ax.text(10, self.material.sigma_Dm1C + 15, r"$\sigma_{D-1}^{C}$")
        ax.text(self.material.sigma_y, 20, r"$\sigma_{y}$")
        ax.text(10, self.material.sigma_y + 15, r"$\sigma_{y}$")
        ax.text(self.material.sigma_u, 20, r"$\sigma_{u}$")
        ax.text(10, P_y - 35, r"$\sigma_{D, lim}$")
        ax.set_xlabel(r"Mean Stress - $\sigma_{m, eq}$ [MPa]")
        ax.set_ylabel(r"Alternating Stress - $\sigma_{a, eq}$ [MPa]")
        plt.title("Haigh Diagram @ " + self.name)
        plt.grid()
        plt.show()
    
    # Calculate Section Static Safety Factor
    def calculateSectionStaticSafetyFactor(self, profile=None):
        zV = self.loc
        for i in range(len(profile.locs)):
            if zV >= profile.locs[i] and zV < profile.locs[i + 1]:
                self.staticSF = profile.shaft.material.sigma_y / profile.sigma_id[i]
    
    # Calculate Section Mean and Alternating Stresses
    def calculateSectionMeanAlternatingStress(self, profile=None):
        zV = self.loc
        for j in range(len(profile.locs)):
            if zV >= profile.locs[j] and zV < profile.locs[j + 1]:
                self.sigma_m_N = profile.sigma_N[j]                    
                self.sigma_a_N = 0                    
                self.sigma_m_Mb = 0                    
                self.sigma_a_Mb = profile.sigma_Mb[j]                    
                self.tau_m_Mt = profile.tau_Mt[j]
                self.tau_a_Mt = 0
    
    # Calculate Fatigue Safety Factor
    def calculateSectionFatigueSafetyFactor(self):
        # Calculate coordinates of specific points
        coeff1 = np.polyfit(np.array([0, self.material.sigma_y]), np.array([self.material.sigma_y, 0]), deg=1)
        coeff2 = np.polyfit(np.array([0, self.material.sigma_u]), np.array([self.material.sigma_Dm1C, 0]), deg=1)
        inter_x = (coeff2[1] - coeff1[1]) / (coeff1[0] - coeff2[0])
        coeff3 = np.polyfit(np.array([0, self.sigma_m_eq]), np.array([0, self.sigma_a_eq]), deg=1)
        P_x = (coeff2[1] - coeff3[1]) / (coeff3[0] - coeff2[0])
        if P_x > inter_x:
            P_x = (coeff1[1] - coeff3[1]) / (coeff3[0] - coeff1[0])
        P_y = np.polyval(coeff3, P_x)
        self.fatigueSF = P_y / self.sigma_a_eq


import numpy as np
'''
This is the "Component" class.
It's a parent class that defines general properties that are common among different components.
Other classes will inherit these properties instead of having to define them every single time.

The properties are:
--> 1) "name": a string of characters acting as a label
--> 2) "material": a "Material" object of the material class defining the material properties of the component
--> 3) "axis": a 3-element vector representing the axis along which the component is rotating with respect to a defined reference frame
--> 4) "loc": a scalar representing the location of the component with respect to the axis of a Shaft component expressed in [mm]
--> 5) "rel_loc": a 3-element vector representing the relative location of the component with respect to the axis of a Shaft component expressed in [mm]
--> 6) "abs_loc": a 3-element vector representing the absolute location of the component with respect to (0, 0, 0) reference frame expressed in [mm]
--> 7) "EFs": a numpy array representing the list of external forces acting on the component expressed in  [N] 
--> 8) "ETs": a numpy array representing the list of external torques acting on the component expressed in [Nm]
--> 9) "onShaft": a Shaft object representing the specific shaft the component is mounted on

These properites can be used and manipulated via the following fucntions:
--> 1) "checkForceEquilibrium(self)": sums the forces in 'EFs' and returns a boolean indicating whether sum is equal to [0, 0, 0]; thsu, satisfying force equilibrium condition.
--> 2) "checkTorqueEquilibrium(self)": virtual function to be implemented where necessary in a child class. It must ensure torque equilibrium on the component.
--> 3) "solve(self)": virtual function to be implemented where necessary in a child class. It must resolve any missing forces/torques acting on the component.
--> 4) "updateEFs(self, EFs=[])": updates the list of external forces "self.EFs" by appending the forces in the input parameter "EFs" if not already in the list.
--> 5) "updateETs(self, ETs=[])": updates the list of external torques "self.ETs" by appending the torques in the input parameter "ETs" if not already in the list.
--> 6) "updateLoc(self)": updates either "self.rel_loc" or "self.abs_loc" based on which one the user gave as input.
'''
class Component:

    # Constructor
    def __init__(self, name="", material=None, axis=np.zeros(3), loc=0, EFs=np.array([]), ETs=np.array([]), omega=np.zeros(3)):
        self.name = name
        self.material = material
        self.axis = axis
        if not isinstance(loc, list):
            self.rel_loc = loc * self.axis
            self.abs_loc = np.array([])
        else:
            self.rel_loc = np.array([])
            self.abs_loc = np.array(loc)
        self.EFs = EFs
        self.ETs = ETs
        self.omega = omega
        self.onShaft = None
    
    # Check force equilibrium
    def checkForceEquilibrium(self):
        print(f"Checking force equilibrium on {self.name}.")
        valid = False
        if self.EFs.size != 0:
            valid = True
        else:
            return valid
        eq = np.zeros(3)
        eqState = False
        for EF in self.EFs:
            eq += EF.force
        if all(np.abs(eq) <= 1e-3 * np.ones(3)):
            print(f"{self.name} maintains a force equilibrium.")
            eqState = True
        else:
            print(f"{self.name} does not maintain a force equilibrium.")
        return eqState

    # Check torque equilibrium
    def checkTorqueEquilibrium(self):
        pass

    # Solve function
    def solve(self):
        pass
    
    # Update external forces
    def updateEFs(self, EFs=[]):
        for ef in EFs:
            if ef not in self.EFs:
                self.EFs = np.append(self.EFs, ef)
    
    # Update external torques
    def updateETs(self, ETs=[]):
        for et in ETs:
            if et not in self.ETs:
                self.ETs = np.append(self.ETs, et)
    
    # Update location
    def updateLoc(self):
        if self.abs_loc.size == 0:
            self.abs_loc = self.rel_loc + self.onShaft.abs_loc
        elif self.rel_loc.size == 0:
            self.rel_loc = self.abs_loc - self.onShaft.abs_loc



'''
This is the "Material" class.
It is used to define the material properties for the components.

The properties are:
--> 1) "name": a string of characters acting as a label
--> 2) "sigma_u": Ultimate tensile strength expressed in [MPa]
--> 3) "sigma_y": Yielding strength expressed in [MPa]
--> 4) "sigma_Dm1": Material specific fatigue limit expressed in [MPa]
--> 5) "HB": Hardness of the material
--> 6) "sigma_Dm1C": Component specific fatigue limit expressed in [MPa]

The component's specific fatigue limit at a specific section can be calculated via the implemented function:
--> 1) "calculateComponentFatigueLimit(self, section=None): calculates the component fatigue limit at specified section.
'''
class Material:

    # Constructor
    def __init__(self, name="", sigma_u=0, sigma_y=0, sigma_Dm1=0, HB=0):
        self.name = name
        self.sigma_u = sigma_u
        self.sigma_y = sigma_y
        self.sigma_Dm1 = sigma_Dm1
        self.HB = HB
    
    # Component Fatigue Limit
    def calculateComponentFatigueLimit(self, section=None):
        self.sigma_Dm1C = self.sigma_Dm1*section.FLCF.C
        return self


'''
This is the "Gear Mesh" class.
It defines a gear mesh by relying on 2 gear components and the type of mesh.

The properties used are:
I) Given properties
--> 1) "drivingGear": a gear component representing the input gear of the gear mesh
--> 2) "drivenGear": a gear component represnting the output gear of the gear mesh
--> 3) "radiality": a 1/2-element vector where each element is composed of 3-element vector representing the direction along which the centers of the driving gear and the driven gear centers are connected with the teeth contact point
--> 4) "type": a string of characters indicating the gear mesh type: Internal / External

II) Calculated properties
--> 1) "ratio": the gear ratio calculated based on the number of teeth
--> 2) "m_G": the reciprocal of the gear ratio
--> 3) "loc": a 3-element vector representing the location of the gear mesh with respect to the origin [0, 0, 0]
--> 4) "F": a 3-element vector representing the resultant force acting on the driven gear due to the gear mesh object defined
--> 5) "F_t": a 3-element vector representing the tangential force component of the resultant force
--> 6) "F_r": a 3-element vector representing the radial force component of the resultant force
--> 7) "F_a": a 3-element vector representing the axial force component of the resultant force

'''
import numpy as np
from .torque import Torque
from .force import Force
class GearMesh:

    # Constructor
    def __init__(self, name="", drivingGear=None, drivenGear=None, radiality=np.zeros(3), type=""):
        if drivingGear == None or drivenGear == None:
            raise ValueError("Driving or driven gear missing!")
        if drivingGear.m_n != drivenGear.m_n:
            raise Exception("Incompatible Gear Mesh!")
        # Update location of driven gear
        if drivenGear.abs_loc.size == 0:
            drivenGear.abs_loc = drivingGear.abs_loc + radiality * (drivingGear.d + drivenGear.d) / 2
        # Given properties
        self.name = name
        self.drivingGear = drivingGear
        self.drivenGear = drivenGear
        self.radiality = radiality
        self.type = type
        # Calculated properties
        self.ratio = self.drivingGear.z / self.drivenGear.z
        self.m_G = 1 / self.ratio
        sgn = 1 # aassuming self.type = "External"
        if self.type == "External":
            sgn = -1
        self.drivenGear.omega = sgn * self.ratio * self.drivingGear.omega
        #self.drivenGear.T_tot = Torque(-sgn * self.drivingGear.T_tot.torque / self.ratio, self.drivenGear.loc)
        self.loc = self.drivingGear.d / 2 * self.radiality + self.drivingGear.abs_loc
        self.F = Force(np.zeros(3), self.loc) # Resultant Force
        self.F_t = Force(np.zeros(3), self.loc) # Tangential Force
        self.F_r = Force(np.zeros(3), self.loc) # Radial Force
        self.F_a = Force(np.zeros(3), self.loc) # Axial Force
        # Update gear meshes
        self.drivingGear.meshes = np.append(self.drivingGear.meshes, self)
        self.drivenGear.meshes = np.append(self.drivenGear.meshes, self)